==> hestia_earth_models-0.65.0-py3-none-any.whl: hestia_earth/models/hestia/seed_emissions.py <==
"""
Seed Emissions

This model uses the `emissions` recalculated from the current `Cycle` to estimate the `emissions` associated with
producing the `Input` with `termType` = `seed`.

These are the steps:

1. Recalculate all emissions for the Cycle of interest;
2. Group Cycle emissions based on their `inputProductionGroupId` lookup value and sum them up;
3. Divide the sum of each emission group by the average crop yield (from FAOSTAT) to get emissions per kg of product;
4. Multiply by `seed.value` to get total background `emissions` from seed production per ha.
"""
from functools import reduce
from hestia_earth.schema import TermTermType, EmissionMethodTier, SiteSiteType
from hestia_earth.utils.lookup import download_lookup, column_name, get_table_value, extract_grouped_data_closest_date
from hestia_earth.utils.model import filter_list_term_type
from hestia_earth.utils.tools import non_empty_list, flatten, list_sum, safe_parse_float
from hestia_earth.utils.emission import cycle_emissions_in_system_boundary

from hestia_earth.models.log import logRequirements, logShouldRun, log_as_table
from hestia_earth.models.utils.emission import _new_emission
from hestia_earth.models.utils.site import valid_site_type
from hestia_earth.models.utils.cycle import cycle_end_year
from hestia_earth.models.utils.crop import get_crop_grouping_faostat_production, get_landCover_term_id
from hestia_earth.models.utils.completeness import _is_term_type_complete
from hestia_earth.models.utils.blank_node import get_lookup_value
from . import MODEL

REQUIREMENTS = {
    "Cycle": {
        "completeness.product": "True",
        "endDate": "",
        "inputs": [{
            "@type": "Input",
            "term.termType": "seed",
            "value": "> 0",
            "none": {
                "impactAssessment": "",
                "fromCycle": "True",
                "producedInCycle": "True"
            }
        }],
        "products": [{
            "@type": "Product",
            "term.termType": "crop",
            "optional": {
                "economicValueShare": "> 0"
            }
        }],
        "site": {
            "@type": "Site",
            "siteType": ["cropland", "glass or high accessible cover"],
            "country": {"@type": "Term", "termType": "region"}
        },
        "emissions": [{"@type": "Emission"}]
    }
}
RETURNS = {
    "Emission": [{
        "value": "",
        "inputs": "",
        "methodTier": "background"
    }]
}
LOOKUPS = {
    "region-crop-cropGroupingFaostatProduction-yield": "value from cropGroupingFaostatProduction and country",
    "crop": [
        "correspondingSeedTermIds",
        "cropGroupingFaostatProduction",
        "global_economic_value_share",
        "landCoverTermId"
    ],
    "emission": "inputProductionGroupId"
}
MODEL_KEY = 'seed_emissions'
TIER = EmissionMethodTier.BACKGROUND.value


def _emission(term_id: str, value: float, input: dict):
    emission = _new_emission(term_id, MODEL)
    emission['value'] = [round(value, 7)]
    emission['inputs'] = [input]
    emission['methodTier'] = TIER
    return emission


def _run(cycle: dict, economicValueShare: float, total_yield: float, seed_input: dict, grouped_emissions: dict):
    term = seed_input.get('term', {})
    seed_value = list_sum(seed_input.get('value'))
    return [
        _emission(term_id, emission_value * economicValueShare / 100 / total_yield * seed_value, term)
        for term_id, emission_value in grouped_emissions.items()
    ]


def _filter_emissions(cycle: dict):
    required_emission_term_ids = cycle_emissions_in_system_boundary(cycle)

    emissions = [
        {
            'id': i.get('term', {}).get('@id'),
            'group-id': get_lookup_value(i.get('term', {}), LOOKUPS['emission'], model=MODEL, model_key=MODEL_KEY),
            'value': list_sum(i.get('value'), 0)
        }
        for i in cycle.get('emissions', [])
        if all([
            i.get('term', {}).get('@id') in required_emission_term_ids,
            list_sum(i.get('value'), 0) > 0
        ])
    ]
    emission_ids = set([v.get('id') for v in emissions])
    group_ids = set([v.get('group-id') for v in emissions if v.get('group-id')])

    # for each group, get the list of all required emissions
    lookup = download_lookup('emission.csv')
    emissions_per_group = [
        {
            'id': group_id,
            'emissions': list(filter(
                lambda id: id in required_emission_term_ids,
                list(lookup[lookup[column_name('inputProductionGroupId')] == group_id].termid)
            ))
        }
        for group_id in group_ids
    ]
    emissions_per_group = [
        {
            'id': group.get('id'),
            'total-emissions': len(group.get('emissions', [])),
            'included-emissions': len([v in emission_ids for v in group.get('emissions', [])]),
            'missing-emissions': '-'.join([v for v in group.get('emissions', []) if v not in emission_ids])
        }
        for group in emissions_per_group
    ]
    # only keep groups that have all emissions present in the Cycle
    valid_groups = list(filter(
        lambda group: group.get('total-emissions') == group.get('included-emissions'),
        emissions_per_group
    ))
    valid_group_ids = set([v.get('id') for v in valid_groups])

    # finally, only return emissions which groups are valid
    return list(filter(
        lambda emission: emission.get('group-id') in valid_group_ids,
        emissions
    )), emissions_per_group


def _evs(product: dict):
    return safe_parse_float(
        get_lookup_value(product.get('term', {}), 'global_economic_value_share', model=MODEL, model_key=MODEL_KEY)
    ) or product.get('economicValueShare')


def _yield(country_id: str, end_year: int, product: dict):
    grouping = get_crop_grouping_faostat_production(MODEL, product.get('term', {}))
    return safe_parse_float(extract_grouped_data_closest_date(get_table_value(
        download_lookup('region-crop-cropGroupingFaostatProduction-yield.csv'),
        'termid',
        country_id,
        column_name(grouping)
    ), end_year)) or list_sum(product.get('value'))


def _group_seed_inputs(inputs: list):
    def _group_by(group: dict, input: dict):
        term_id = input.get('term', {}).get('@id')
        group[term_id] = group.get(term_id, []) + [input]
        return group

    grouped_inputs = reduce(_group_by, inputs, {})
    return [
        inputs[0] | {'value': flatten([v.get('value') for v in inputs])}
        for inputs in grouped_inputs.values()
    ]


def _should_run(cycle: dict):
    crop_products = filter_list_term_type(cycle.get('products', []), TermTermType.CROP)
    site_type_valid = valid_site_type(cycle.get('site', {}), [
        SiteSiteType.CROPLAND.value, SiteSiteType.GLASS_OR_HIGH_ACCESSIBLE_COVER.value
    ])
    is_product_complete = _is_term_type_complete(cycle, 'product')
    end_year = cycle_end_year(cycle)
    country_id = cycle.get('site', {}).get('country', {}).get('@id')

    # only keep the crop products that map to a seed
    crop_products = [
        {
            'product': product.get('term', {}).get('@id'),
            'seed-id': get_lookup_value(
                product.get('term', {}), 'correspondingSeedTermIds', model=MODEL, key=MODEL_KEY),
            'economicValueShare': _evs(product),
            'yield': _yield(country_id, end_year, product),
            'landCover-id': get_landCover_term_id(product.get('term', {}), model=MODEL, key=MODEL_KEY)
        }
        for product in crop_products
    ]
    valid_crop_products = [
        value for value in crop_products if all([
            value.get('seed-id'),
            value.get('economicValueShare'),
            value.get('yield'),
            value.get('landCover-id'),
        ])
    ]

    # array of ; delimited values
    seed_term_ids = list(set(flatten([v.get('seed-id').split(';') for v in valid_crop_products if v.get('seed-id')])))

    seed_inputs = [
        i
        for i in cycle.get('inputs', [])
        if all([
            i.get('term', {}).get('termType') == TermTermType.SEED.value,
            i.get('term', {}).get('@id') in seed_term_ids,
            list_sum(i.get('value'), 0) > 0,
            not i.get('impactAssessment'),
            # ignore inputs which are flagged as Product of the Cycle
            not i.get('fromCycle', False),
            not i.get('producedInCycle', False)
        ])
    ]
    # sum up seed inputs with the same id
    seed_inputs = _group_seed_inputs(seed_inputs)

    crop_land_cover_ids = list(set([p.get('landCover-id') for p in valid_crop_products]))
    total_economicValueShare = list_sum([p.get('economicValueShare') for p in valid_crop_products])
    total_yield = list_sum([p.get('yield') for p in valid_crop_products])

    emissions, emissions_per_group = _filter_emissions(cycle)
    # group emissions with the same group-id
    grouped_emissions = reduce(
        lambda p, c: p | {c.get('group-id'): p.get(c.get('group-id'), 0) + c.get('value', 0)},
        emissions,
        {}
    )

    should_run = all([
        site_type_valid,
        len(crop_land_cover_ids) <= 1,
        is_product_complete,
        total_economicValueShare,
        total_yield,
        bool(seed_inputs),
        bool(emissions)
    ])

    for seed_input in seed_inputs:
        term_id = seed_input.get('term', {}).get('@id')

        logRequirements(cycle, model=MODEL, term=term_id, model_key=MODEL_KEY,
                        site_type_valid=site_type_valid,
                        crop_products=log_as_table(crop_products),
                        crop_land_cover_ids=';'.join(crop_land_cover_ids),
                        is_term_type_product_complete=is_product_complete,
                        total_economicValueShare=total_economicValueShare,
                        total_yield=total_yield,
                        end_year=end_year,
                        country_id=country_id,
                        emissions=log_as_table(emissions),
                        emissions_per_group=log_as_table(emissions_per_group))

        logShouldRun(cycle, MODEL, term_id, should_run, key=MODEL_KEY)

    return should_run, total_economicValueShare, total_yield, seed_inputs, grouped_emissions


def run(cycle: dict):
    should_run, economicValueShare, total_yield, seed_inputs, grouped_emissions = _should_run(cycle)
    return flatten(non_empty_list([
        _run(cycle, economicValueShare, total_yield, seed_input, grouped_emissions)
        for seed_input in seed_inputs
    ])) if should_run else []

==> hestia_earth_models-0.65.2-py3-none-any.whl: hestia_earth/models/hestia/seed_emissions.py <==
"""
Seed Emissions

This model uses the `emissions` recalculated from the current `Cycle` to estimate the `emissions` associated with
producing the `Input` with `termType` = `seed`.

These are the steps:

1. Recalculate all emissions for the Cycle of interest;
2. Group Cycle emissions based on their `inputProductionGroupId` lookup value and sum them up;
3. Divide the sum of each emission group by the average crop yield (from FAOSTAT) to get emissions per kg of product;
4. Multiply by `seed.value` to get total background `emissions` from seed production per ha.
"""
from functools import reduce
from hestia_earth.schema import TermTermType, EmissionMethodTier, SiteSiteType
from hestia_earth.utils.lookup import download_lookup, column_name, get_table_value, extract_grouped_data_closest_date
from hestia_earth.utils.model import filter_list_term_type
from hestia_earth.utils.tools import non_empty_list, flatten, list_sum, safe_parse_float
from hestia_earth.utils.emission import cycle_emissions_in_system_boundary

from hestia_earth.models.log import logRequirements, logShouldRun, log_as_table
from hestia_earth.models.utils import _omit
from hestia_earth.models.utils.emission import _new_emission
from hestia_earth.models.utils.site import valid_site_type
from hestia_earth.models.utils.cycle import cycle_end_year
from hestia_earth.models.utils.crop import get_crop_grouping_faostat_production, get_landCover_term_id
from hestia_earth.models.utils.completeness import _is_term_type_complete
from hestia_earth.models.utils.blank_node import get_lookup_value
from . import MODEL

REQUIREMENTS = {
    "Cycle": {
        "completeness.product": "True",
        "endDate": "",
        "inputs": [{
            "@type": "Input",
            "term.termType": "seed",
            "value": "> 0",
            "none": {
                "impactAssessment": "",
                "fromCycle": "True",
                "producedInCycle": "True"
            }
        }],
        "products": [{
            "@type": "Product",
            "term.termType": "crop",
            "optional": {
                "economicValueShare": "> 0"
            }
        }],
        "site": {
            "@type": "Site",
            "siteType": ["cropland", "glass or high accessible cover"],
            "country": {"@type": "Term", "termType": "region"}
        },
        "emissions": [{"@type": "Emission"}]
    }
}
RETURNS = {
    "Emission": [{
        "value": "",
        "inputs": "",
        "methodTier": "background"
    }]
}
LOOKUPS = {
    "region-crop-cropGroupingFaostatProduction-yield": "value from cropGroupingFaostatProduction and country",
    "crop": [
        "correspondingSeedTermIds",
        "cropGroupingFaostatProduction",
        "global_economic_value_share",
        "landCoverTermId"
    ],
    "emission": "inputProductionGroupId"
}
MODEL_KEY = 'seed_emissions'
TIER = EmissionMethodTier.BACKGROUND.value


def _emission(term_id: str, value: float, input: dict):
    emission = _new_emission(term_id, MODEL)
    emission['value'] = [round(value, 7)]
    emission['inputs'] = [input]
    emission['methodTier'] = TIER
    return emission


def _run(cycle: dict, economicValueShare: float, total_yield: float, seed_input: dict, grouped_emissions: dict):
    term = seed_input.get('term', {})
    seed_value = list_sum(seed_input.get('value'))
    return [
        _emission(term_id, emission_value * economicValueShare / 100 / total_yield * seed_value, term)
        for term_id, emission_value in grouped_emissions.items()
    ]


def _filter_emissions(cycle: dict):
    required_emission_term_ids = cycle_emissions_in_system_boundary(cycle)

    emissions = [
        {
            'id': i.get('term', {}).get('@id'),
            'group-id': get_lookup_value(i.get('term', {}), LOOKUPS['emission'], model=MODEL, model_key=MODEL_KEY),
            'value': list_sum(i.get('value'), 0)
        }
        for i in cycle.get('emissions', [])
        if all([
            i.get('term', {}).get('@id') in required_emission_term_ids,
            list_sum(i.get('value'), 0) > 0
        ])
    ]
    emission_ids = set([v.get('id') for v in emissions])
    group_ids = set([v.get('group-id') for v in emissions if v.get('group-id')])

    # for each group, get the list of all required emissions
    lookup = download_lookup('emission.csv')
    emissions_per_group = [
        {
            'id': group_id,
            'emissions': list(filter(
                lambda id: id in required_emission_term_ids,
                list(lookup[lookup[column_name('inputProductionGroupId')] == group_id].termid)
            ))
        }
        for group_id in group_ids
    ]
    emissions_per_group = [
        {
            'id': group.get('id'),
            'total-emissions': len(group.get('emissions', [])),
            'included-emissions': len([v in emission_ids for v in group.get('emissions', [])]),
            'missing-emissions': '-'.join([v for v in group.get('emissions', []) if v not in emission_ids])
        }
        for group in emissions_per_group
    ]
    # only keep groups that have all emissions present in the Cycle
    valid_groups = list(filter(
        lambda group: group.get('total-emissions') == group.get('included-emissions'),
        emissions_per_group
    ))
    valid_group_ids = set([v.get('id') for v in valid_groups])

    # finally, only return emissions which groups are valid
    return list(filter(
        lambda emission: emission.get('group-id') in valid_group_ids,
        emissions
    )), emissions_per_group


def _evs(product: dict):
    return safe_parse_float(
        get_lookup_value(product.get('term', {}), 'global_economic_value_share', model=MODEL, model_key=MODEL_KEY)
    ) or product.get('economicValueShare')


def _yield(country_id: str, end_year: int, product: dict):
    grouping = get_crop_grouping_faostat_production(MODEL, product.get('term', {}))
    return safe_parse_float(extract_grouped_data_closest_date(get_table_value(
        download_lookup('region-crop-cropGroupingFaostatProduction-yield.csv'),
        'termid',
        country_id,
        column_name(grouping)
    ), end_year)) or list_sum(product.get('value'))


def _group_seed_inputs(inputs: list):
    def _group_by(group: dict, input: dict):
        term_id = input.get('term', {}).get('@id')
        group[term_id] = group.get(term_id, []) + [input]
        return group

    grouped_inputs = reduce(_group_by, inputs, {})
    return [
        inputs[0] | {'value': flatten([v.get('value') for v in inputs])}
        for inputs in grouped_inputs.values()
    ]


def _should_run(cycle: dict):
    crop_products = filter_list_term_type(cycle.get('products', []), TermTermType.CROP)
    site_type_valid = valid_site_type(cycle.get('site', {}), [
        SiteSiteType.CROPLAND.value, SiteSiteType.GLASS_OR_HIGH_ACCESSIBLE_COVER.value
    ])
    is_product_complete = _is_term_type_complete(cycle, 'product')
    end_year = cycle_end_year(cycle)
    country_id = cycle.get('site', {}).get('country', {}).get('@id')

    # only keep the crop products that map to a seed
    crop_products = [
        {
            'product': product.get('term', {}).get('@id'),
            'seed-id': get_lookup_value(
                product.get('term', {}), 'correspondingSeedTermIds', model=MODEL, key=MODEL_KEY) or None,
            'economicValueShare': _evs(product),
            'yield': _yield(country_id, end_year, product),
            'landCover-id': get_landCover_term_id(product.get('term', {}), model=MODEL, key=MODEL_KEY)
        }
        for product in crop_products
    ]
    valid_crop_products = [
        value for value in crop_products if all([
            value.get('seed-id'),
            value.get('economicValueShare'),
            value.get('yield'),
            value.get('landCover-id'),
        ])
    ]

    # array of ; delimited values
    seed_term_ids = list(set(flatten([v.get('seed-id').split(';') for v in valid_crop_products if v.get('seed-id')])))

    seed_inputs = [
        {
            'input': i,
            'is-corresponding-seed': i.get('term', {}).get('@id') in seed_term_ids,
            'input-value': i.get('value'),
            'has-linked-impact-assessment': bool(i.get('impactAssessment')),
            'is-fromCycle': i.get('fromCycle', False),
            'is-producedInCycle': i.get('producedInCycle', False),
        }
        for i in filter_list_term_type(cycle.get('inputs', []), TermTermType.SEED)
    ]
    # sum up seed inputs with the same id
    grouped_seed_inputs = _group_seed_inputs([
        v.get('input') for v in seed_inputs
        if all([
            v.get('is-corresponding-seed', False),
            list_sum(v.get('input-value') or [-1]) > 0,
            not v.get('has-linked-impact-assessment'),
            not v.get('is-fromCycle'),
            not v.get('is-producedInCycle'),
        ])
    ])

    crop_land_cover_ids = list(set([p.get('landCover-id') for p in valid_crop_products]))
    total_economicValueShare = list_sum([p.get('economicValueShare') for p in valid_crop_products])
    total_yield = list_sum([p.get('yield') for p in valid_crop_products])

    emissions, emissions_per_group = _filter_emissions(cycle)
    # group emissions with the same group-id
    grouped_emissions = reduce(
        lambda p, c: p | {c.get('group-id'): p.get(c.get('group-id'), 0) + c.get('value', 0)},
        emissions,
        {}
    )

    should_run = all([
        site_type_valid,
        len(crop_land_cover_ids) <= 1,
        is_product_complete,
        total_economicValueShare,
        total_yield,
        bool(grouped_seed_inputs),
        bool(emissions)
    ])

    for seed_input in seed_inputs:
        term_id = seed_input.get('input').get('term', {}).get('@id')

        logRequirements(cycle, model=MODEL, term=term_id, model_key=MODEL_KEY,
                        site_type_valid=site_type_valid,
                        crop_products=log_as_table(crop_products),
                        crop_land_cover_ids=';'.join(crop_land_cover_ids),
                        is_term_type_product_complete=is_product_complete,
                        total_economicValueShare=total_economicValueShare,
                        total_yield=total_yield,
                        end_year=end_year,
                        country_id=country_id,
                        emissions=log_as_table(emissions),
                        emissions_per_group=log_as_table(emissions_per_group),
                        **_omit(seed_input, 'input'))

        logShouldRun(cycle, MODEL, term_id, should_run, key=MODEL_KEY)

    return should_run, total_economicValueShare, total_yield, grouped_seed_inputs, grouped_emissions


def run(cycle: dict):
    should_run, economicValueShare, total_yield, seed_inputs, grouped_emissions = _should_run(cycle)
    return flatten(non_empty_list([
        _run(cycle, economicValueShare, total_yield, seed_input, grouped_emissions)
        for seed_input in seed_inputs
    ])) if should_run else []
